@@ -11,8 +11,10 @@
 //! crate.
 
 use crate::internals::zeroize as internals;
-use crate::macros::precondition_memory_range;
-use mirai_annotations::{debug_checked_precondition, debug_checked_precondition_eq};
+use crate::macros::{
+    debug_precondition_logaligned, debug_precondition_logmultiple, precondition_memory_range,
+};
+use crate::util::is_aligned_ptr_mut;
 
 /// Strategy for securely erasing memory.
 ///
@@ -28,37 +30,33 @@ pub trait MemZeroizer {
     /// The caller *must* ensure that `ptr` is valid for writes of `len` bytes,
     /// see the [`std::ptr`] documentation. In particular this function is
     /// not atomic.
-    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize);
+    ///
+    /// Furthermore, `ptr` *must* be at least `2^LOG_ALIGN` byte aligned, and
+    /// `2^LOG_ALIGN` must fit a `usize`.
+    ///
+    /// Finally `len` must be a multiple of `2^LOG_MULTIPLE`, and `2^LOG_ALIGN`
+    /// must fit a `usize`. (If `len` is not a multiple of `2^LOG_MULTIPLE`
+    /// then this won't result in UB but the memory pointed to by `ptr` might
+    /// only be zeroized for `len` rounded down to a multiple `2^LOG_MULTIPLE`
+    /// bytes, or the full `len` bytes, or anything in between, or the function
+    /// might panic.)
+    unsafe fn zeroize_mem_blocks<const LOG_ALIGN: u8, const LOG_MULTIPLE: u8>(
+        &self,
+        ptr: *mut u8,
+        len: usize,
+    );
 
-    /// Zeroize the memory pointed to by `ptr` and of size `len` bytes, aligned
-    /// at least `align`.
+    /// Zeroize the memory pointed to by `ptr` and of size `len` bytes.
+    /// Shorthand for `Self::zeroize_mem_blocks::<0, 0>`.
     ///
     /// This is guarantied to be not elided by the compiler.
     ///
-    /// `ptr` must be at least `align` byte aligned, see the safety section
-    /// below. The `align` value might be used to optimise out a branch on
-    /// alignment if `align` is known at compile time.
-    ///
     /// # Safety
     /// The caller *must* ensure that `ptr` is valid for writes of `len` bytes,
     /// see the [`std::ptr`] documentation. In particular this function is
     /// not atomic.
-    ///
-    /// Furthermore, `ptr` *must* be at least `align` byte aligned, and `align`
-    /// must be a power of 2 (and therefore non-zero).
-    ///
-    /// # Performance
-    /// The `align` value might be used to optimise out a branch on alignment if
-    /// `align` is known at compile time. Using this method will at least
-    /// not degrade performance relative to [`Self::zeroize_mem`] if `align` is
-    /// known at compile time. Therefore it is fine to underestimate the
-    /// alignment, especially if this underestimate can be known at compile
-    /// time.
-    unsafe fn zeroize_mem_minaligned(&self, ptr: *mut u8, len: usize, _align: usize) {
-        precondition_memory_range!(ptr, len);
-        // SAFETY: caller must uphold the safety contract of `self.zeroize_mem` (and
-        // more)
-        unsafe { self.zeroize_mem(ptr, len) }
+    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize) {
+        unsafe { self.zeroize_mem_blocks::<0, 0>(ptr, len) }
     }
 }
 
@@ -90,6 +88,20 @@ cfg_if::cfg_if! {
         /// the selected features and the version of this library.
         pub type DefaultMemZeroizer = LibcZeroizer;
         pub(crate) use LibcZeroizer as DefaultMemZeroizerConstructor;
+    } else if #[cfg(all(target_arch = "x86_64", target_feature = "avx"))] {
+        /// Best (i.e. fastest) [`MemZeroizer`] available for the target.
+        ///
+        /// Which [`MemZeroizer`] this is is an implementation detail, can depend on the target and
+        /// the selected features and the version of this library.
+        pub type DefaultMemZeroizer = X86_64AvxZeroizer;
+        pub(crate) use X86_64AvxZeroizer as DefaultMemZeroizerConstructor;
+    } else if #[cfg(all(target_arch = "x86_64", target_feature = "sse2"))] {
+        /// Best (i.e. fastest) [`MemZeroizer`] available for the target.
+        ///
+        /// Which [`MemZeroizer`] this is is an implementation detail, can depend on the target and
+        /// the selected features and the version of this library.
+        pub type DefaultMemZeroizer = X86_64Sse2Zeroizer;
+        pub(crate) use X86_64Sse2Zeroizer as DefaultMemZeroizerConstructor;
     } else {
         /// Best (i.e. fastest) [`MemZeroizer`] available for the target.
         ///
@@ -118,8 +130,9 @@ pub struct VolatileMemsetZeroizer;
 
 #[cfg(feature = "nightly_core_intrinsics")]
 impl MemZeroizer for VolatileMemsetZeroizer {
-    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize) {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, ptr: *mut u8, len: usize) {
         precondition_memory_range!(ptr, len);
+        debug_precondition_logaligned!(A, ptr);
         // SAFETY: the caller must uphold the safety contract of
         // `internals::volatile_memset`
         unsafe {
@@ -161,8 +174,10 @@ pub struct LibcZeroizer;
     target_env = "musl"
 ))]
 impl MemZeroizer for LibcZeroizer {
-    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize) {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, ptr: *mut u8, len: usize) {
         precondition_memory_range!(ptr, len);
+        debug_precondition_logaligned!(A, ptr);
+        debug_precondition_logmultiple!(B, len);
         // SAFETY: the caller must uphold the safety contract of
         // `internals::libc_explicit_bzero`
         unsafe {
@@ -187,12 +202,14 @@ pub struct AsmRepStosZeroizer;
 
 #[cfg(all(target_arch = "x86_64", target_feature = "ermsb"))]
 impl MemZeroizer for AsmRepStosZeroizer {
-    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize) {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, ptr: *mut u8, len: usize) {
         precondition_memory_range!(ptr, len);
+        debug_precondition_logaligned!(A, ptr);
+        debug_precondition_logmultiple!(B, len);
         // SAFETY: the caller must uphold the safety contract of
-        // `internals::c_asm_ermsb_zeroize`
+        // `internals::asm_ermsb_zeroize`
         unsafe {
-            internals::c_asm_ermsb_zeroize(ptr, len);
+            internals::asm_ermsb_zeroize(ptr, len);
         }
         fence();
     }
@@ -211,12 +228,14 @@ impl MemZeroizer for AsmRepStosZeroizer {
 pub struct VolatileWriteZeroizer;
 
 impl MemZeroizer for VolatileWriteZeroizer {
-    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize) {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, ptr: *mut u8, len: usize) {
         precondition_memory_range!(ptr, len);
+        debug_precondition_logaligned!(A, ptr);
+        debug_precondition_logmultiple!(B, len);
         // SAFETY: the caller must uphold the safety contract of
         // `volatile_write_zeroize_mem`
         unsafe {
-            volatile_write_zeroize_mem(ptr, len);
+            internals::volatile_write_zeroize(ptr, len);
         }
         fence();
     }
@@ -234,127 +253,140 @@ impl MemZeroizer for VolatileWriteZeroizer {
 /// The use of this fence is therefore only a precaution.
 ///
 /// This zeroization method can benefit (in terms of performance) from using the
-/// [`MemZeroizer::zeroize_mem_minaligned`] function instead of
-/// [`MemZeroizer::zeroize_mem`] function if a minimum alignment might be known
+/// [`MemZeroizer::zeroize_mem_blocks`] function instead of
+/// [`MemZeroizer::zeroize_mem`] function if a minimum alignment is known
 /// at compile time.
 #[derive(Debug, Copy, Clone, Default)]
 pub struct VolatileWrite8Zeroizer;
 
 impl MemZeroizer for VolatileWrite8Zeroizer {
-    unsafe fn zeroize_mem_minaligned(&self, ptr: *mut u8, len: usize, align: usize) {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, mut ptr: *mut u8, len: usize) {
         precondition_memory_range!(ptr, len);
-        debug_checked_precondition!(align.is_power_of_two());
-        debug_checked_precondition_eq!((ptr as usize) % align, 0);
-        // if we have 8 byte alignment then write 8 bytes at a time, otherwise
-        // byte-for-byte
-        if align >= 8 {
+        debug_precondition_logaligned!(A, ptr);
+        debug_precondition_logmultiple!(B, len);
+        // if we have 8 = 2^3 byte alignment then write 8 bytes at a time,
+        // otherwise byte-for-byte
+        if (A >= 3) | is_aligned_ptr_mut(ptr, 8) {
             // SAFETY: by the above check, `ptr` is at least 8 byte aligned
-            // SAFETY: the other safety requirements of `volatile_write8_zeroize_mem` are
-            // also required by this function
-            unsafe {
-                volatile_write8_zeroize_mem(ptr, len);
+            // SAFETY: the other safety requirements uphold by caller
+            ptr = unsafe { internals::zeroize_align8_block8(ptr, len) };
+            if B < 3 {
+                unsafe { internals::zeroize_align4_tail8(ptr, len) };
             }
         } else {
             // SAFETY: the caller must uphold the contract of `volatile_write_zeroize_mem`
             unsafe {
-                self.zeroize_mem(ptr, len);
-            }
-        }
-        fence();
-    }
-
-    unsafe fn zeroize_mem(&self, ptr: *mut u8, len: usize) {
-        precondition_memory_range!(ptr, len);
-        if (ptr as usize) % 8 == 0 {
-            // SAFETY: by the above check, `ptr` is at least 8 byte aligned
-            // SAFETY: the other safety requirements of `volatile_write8_zeroize_mem` are
-            // also required by this function
-            unsafe {
-                volatile_write8_zeroize_mem(ptr, len);
-            }
-        } else {
-            // SAFETY: the caller must uphold the contract of `volatile_write_zeroize_mem`
-            unsafe {
-                volatile_write_zeroize_mem(ptr, len);
+                internals::volatile_write_zeroize(ptr, len);
             }
         }
         fence();
     }
 }
 
-/// Zeroize the memory pointed to by `ptr` and of size `len` bytes, by
-/// overwriting it byte for byte using volatile writes.
+/// This zeroizer uses inline asm with avx2 instructions if the pointer is 32
+/// byte aligned, and otherwise uses `VolatileWrite8Zeroizer`. This zeroization
+/// technique is available for x86_64 platforms with avx2 cpu support on stable,
+/// and reasonably fast for 32 byte aligned pointers.
 ///
-/// This is guarantied to be not elided by the compiler.
+/// In addition to the volatile write we place a compiler fence right next to
+/// the volatile write. This should not be necessary for secure zeroization
+/// since the volatile semantics guarenties our writes are not elided, and they
+/// can not be delayed since we are deallocating the memory after zeroization.
+/// The use of this fence is therefore only a precaution.
 ///
-/// # Safety
-/// The caller *must* ensure that `ptr` is valid for writes of `len` bytes, see
-/// the [`std::ptr`] documentation. In particular this function is not atomic.
-unsafe fn volatile_write_zeroize_mem(mut ptr: *mut u8, len: usize) {
-    precondition_memory_range!(ptr, len);
-    for _i in 0..len {
-        // SAFETY: `ptr` originally pointed into an allocation of `len` bytes so now,
-        // after `_i` steps `len - _i > 0` bytes are left, so `ptr` is valid for
-        // a byte write
-        unsafe {
-            core::ptr::write_volatile(ptr, 0u8);
+/// This zeroization method can benefit (in terms of performance) from using the
+/// [`MemZeroizer::zeroize_mem_blocks`] function instead of
+/// [`MemZeroizer::zeroize_mem`] function if a minimum alignment is known
+/// at compile time.
+#[cfg(all(target_arch = "x86_64", target_feature = "avx"))]
+#[derive(Debug, Copy, Clone, Default)]
+pub struct X86_64AvxZeroizer;
+
+#[cfg(all(target_arch = "x86_64", target_feature = "avx"))]
+impl MemZeroizer for X86_64AvxZeroizer {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, mut ptr: *mut u8, len: usize) {
+        precondition_memory_range!(ptr, len);
+        debug_precondition_logaligned!(A, ptr);
+        debug_precondition_logmultiple!(B, len);
+        // if we have 32 = 2^5 byte alignment then write 32 bytes at a time,
+        // with 8 = 2^3 byte align do 8 bytes at a time, otherwise 1 byte at a time
+        if (A >= 5) | is_aligned_ptr_mut(ptr, 32) {
+            // SAFETY: `ptr` is 32 byte aligned
+            ptr = unsafe { internals::x86_64_simd32_unroll2_zeroize_align32_block32(ptr, len) };
+            // zeroize tail
+            if B < 5 {
+                ptr = unsafe { internals::zeroize_align8_block8(ptr, len % 32) };
+            }
+            if B < 3 {
+                unsafe { internals::zeroize_align4_tail8(ptr, len % 8) };
+            }
+        } else if (A >= 3) | is_aligned_ptr_mut(ptr, 8) {
+            // SAFETY: `ptr` is 8 byte aligned
+            ptr = unsafe { internals::zeroize_align8_block8(ptr, len % 32) };
+            if B < 3 {
+                unsafe { internals::zeroize_align4_tail8(ptr, len % 8) };
+            }
+        } else {
+            // SAFETY: no alignment requirement
+            unsafe {
+                internals::volatile_write_zeroize(ptr, len);
+            }
         }
-        // SAFETY: after increment, `ptr` points into the same allocation if `_i == len`
-        // or one byte past it, so `add` is sound
-        ptr = unsafe { ptr.add(1) };
+        fence();
     }
 }
 
-/// Zeroize the memory pointed to by `ptr` and of size `len` bytes, by
-/// overwriting it 8 bytes at a time using volatile writes.
+/// This zeroizer uses inline asm with sse2 instructions if the pointer is 16
+/// byte aligned, and otherwise uses `VolatileWrite8Zeroizer`. This zeroization
+/// technique is available for x86_64 platforms with sse2 cpu support on stable,
+/// and reasonably fast for 16 byte aligned pointers.
 ///
-/// This is guarantied to be not elided by the compiler.
+/// In addition to the volatile write we place a compiler fence right next to
+/// the volatile write. This should not be necessary for secure zeroization
+/// since the volatile semantics guarenties our writes are not elided, and they
+/// can not be delayed since we are deallocating the memory after zeroization.
+/// The use of this fence is therefore only a precaution.
 ///
-/// # Safety
-/// The caller *must* ensure that `ptr` is valid for writes of `len` bytes, see
-/// the [`std::ptr`] documentation. In particular this function is not atomic.
-///
-/// Furthermore, `ptr` *must* be at least 8 byte aligned.
-unsafe fn volatile_write8_zeroize_mem(mut ptr: *mut u8, len: usize) {
-    precondition_memory_range!(ptr, len);
-    debug_checked_precondition_eq!((ptr as usize) % 8, 0);
-    let nblocks = (len - len % 8) / 8;
-    for _i in 0..nblocks {
-        // SAFETY: `ptr` originally pointed into an allocation of `len` bytes so now,
-        // after `_i` steps `len - 8*_i >= 8` bytes are left, so `ptr` is valid
-        // for an 8 byte write SAFETY: `ptr` was originally 8 byte aligned by
-        // caller contract and we only added a multiple of 8 so it is still 8
-        // byte aligned
-        unsafe {
-            core::ptr::write_volatile(ptr.cast::<u64>(), 0u64);
+/// This zeroization method can benefit (in terms of performance) from using the
+/// [`MemZeroizer::zeroize_mem_blocks`] function instead of
+/// [`MemZeroizer::zeroize_mem`] function if a minimum alignment is known
+/// at compile time.
+#[cfg(all(target_arch = "x86_64", target_feature = "sse2"))]
+#[derive(Debug, Copy, Clone, Default)]
+pub struct X86_64Sse2Zeroizer;
+
+#[cfg(all(target_arch = "x86_64", target_feature = "sse2"))]
+impl MemZeroizer for X86_64Sse2Zeroizer {
+    unsafe fn zeroize_mem_blocks<const A: u8, const B: u8>(&self, mut ptr: *mut u8, len: usize) {
+        precondition_memory_range!(ptr, len);
+        debug_precondition_logaligned!(A, ptr);
+        debug_precondition_logmultiple!(B, len);
+        // if we have 16 = 2^4 byte alignment then write 16 bytes at a time,
+        // with 8 = 2^3 byte align do 8 bytes at a time, otherwise 1 byte at a time
+        if (A >= 4) | is_aligned_ptr_mut(ptr, 16) {
+            // SAFETY: `ptr` is 16 byte aligned
+
+            ptr = unsafe { internals::x86_64_simd16_unroll2_zeroize_align16_block16(ptr, len) };
+            // zeroize tail
+            if B < 4 {
+                ptr = unsafe { internals::zeroize_align8_block8(ptr, len % 16) };
+            }
+            if B < 3 {
+                unsafe { internals::zeroize_align4_tail8(ptr, len % 8) };
+            }
+        } else if (A >= 3) | is_aligned_ptr_mut(ptr, 8) {
+            // SAFETY: `ptr` is 8 byte aligned
+            ptr = unsafe { internals::zeroize_align8_block8(ptr, len % 16) };
+            if B < 3 {
+                unsafe { internals::zeroize_align4_tail8(ptr, len % 8) };
+            }
+        } else {
+            // SAFETY: no alignment requirement
+            unsafe {
+                internals::volatile_write_zeroize(ptr, len);
+            }
         }
-        // SAFETY: after increment, `ptr` points into the same allocation or (if `8*_i
-        // == len`) at most one byte past it, so `add` is sound; `ptr` stays 8
-        // byte aligned
-        ptr = unsafe { ptr.add(8) };
-    }
-    // if `len` is not a multiple of 8 then the remainder (at most 7 bytes) needs to
-    // be zeroized; if the remainder is at least 4 bytes we zero these with a single
-    // write
-    if len % 8 >= 4 {
-        // SAFETY: `ptr` has been incremented by a multiple of 8 <= `len` so `ptr`
-        // points to an allocation of `len % 8 >= 4` bytes, so `ptr` is valid
-        // for a 4 byte write SAFETY: `ptr` is still 8 byte aligned so also 4
-        // byte aligned
-        unsafe {
-            core::ptr::write_volatile(ptr.cast::<u32>(), 0u32);
-        }
-        ptr = unsafe { ptr.add(4) };
-    }
-    // the final remainder (at most 3 bytes) is zeroed byte-for-byte
-    // SAFETY: `ptr` has been incremented by a multiple of 4 <= `len` so `ptr`
-    // points to an allocation of `len % 4` bytes, so `ptr` can be written to
-    // and incremented `len % 4` times
-    for _i in 0..(len % 4) {
-        unsafe {
-            core::ptr::write_volatile(ptr, 0u8);
-        }
-        ptr = unsafe { ptr.add(1) };
+        fence();
     }
 }
 
