@@ -49,7 +49,7 @@ fn test_b127_libc_zeroizer() {
     test_b127_zeroizer(LibcZeroizer);
 }
 
-#[cfg(all(target_arch = "x86_64", target_feature = "ermsb", feature = "cc"))]
+#[cfg(all(target_arch = "x86_64", target_feature = "ermsb"))]
 #[test]
 #[cfg_attr(miri, ignore)] // ffi, asm
 fn test_b127_asm_rep_stos_zeroizer() {
@@ -66,6 +66,18 @@ fn test_b127_volatile_write8_zeroizer() {
     test_b127_zeroizer(VolatileWrite8Zeroizer);
 }
 
+#[cfg(all(target_arch = "x86_64", target_feature = "avx"))]
+#[test]
+fn test_b127_x86_64_avx_zeroizer() {
+    test_b127_zeroizer(X86_64AvxZeroizer);
+}
+
+#[cfg(all(target_arch = "x86_64", target_feature = "sse2"))]
+#[test]
+fn test_b127_x86_64_sse2_zeroizer() {
+    test_b127_zeroizer(X86_64Sse2Zeroizer);
+}
+
 #[cfg(feature = "nightly_core_intrinsics")]
 #[test]
 fn test_b239_lowalign_volatile_memset_zeroizer() {
@@ -88,7 +100,7 @@ fn test_b239_lowalign_libc_zeroizer() {
     test_b239_lowalign_zeroizer(LibcZeroizer);
 }
 
-#[cfg(all(target_arch = "x86_64", target_feature = "ermsb", feature = "cc"))]
+#[cfg(all(target_arch = "x86_64", target_feature = "ermsb"))]
 #[test]
 #[cfg_attr(miri, ignore)] // ffi, asm
 fn test_b239_lowalign_asm_rep_stos_zeroizer() {
@@ -104,3 +116,15 @@ fn test_b239_lowalign_volatile_write_zeroizer() {
 fn test_b239_lowalign_volatile_write8_zeroizer() {
     test_b239_lowalign_zeroizer(VolatileWrite8Zeroizer);
 }
+
+#[cfg(all(target_arch = "x86_64", target_feature = "avx"))]
+#[test]
+fn test_b239_lowalign_x86_64_avx_zeroizer() {
+    test_b239_lowalign_zeroizer(X86_64AvxZeroizer);
+}
+
+#[cfg(all(target_arch = "x86_64", target_feature = "sse2"))]
+#[test]
+fn test_b239_lowalign_x86_64_sse2_zeroizer() {
+    test_b239_lowalign_zeroizer(X86_64Sse2Zeroizer);
+}
