@@ -18,6 +18,8 @@ use crate::zeroize::{DefaultMemZeroizer, DefaultMemZeroizerConstructor, MemZeroi
 use alloc::alloc::handle_alloc_error;
 use core::alloc::{GlobalAlloc, Layout};
 use core::ptr::NonNull;
+#[cfg(not(feature = "nightly_strict_provenance"))]
+use sptr::Strict;
 
 /// Wrapper around an allocator which zeroizes memory on deallocation. See the
 /// module level documentation.
@@ -93,7 +95,7 @@ where
 
         if cfg!(debug_assertions) {
             // you can't wrap around the address space
-            if (ptr as usize).checked_add(layout.size()).is_none() {
+            if ptr.addr().checked_add(layout.size()).is_none() {
                 handle_alloc_error(layout);
             }
         }
@@ -104,8 +106,7 @@ where
         // and not yet deallocated SAFETY: `ptr` is at least `layout.align()`
         // byte aligned and this is a power of two
         unsafe {
-            self.zeroizer
-                .zeroize_mem_minaligned(ptr, layout.size(), layout.align());
+            self.zeroizer.zeroize_mem(ptr, layout.size());
         }
         // SAFETY: caller must uphold the safety contract of `GlobalAlloc::dealloc`.
         unsafe { self.backend_alloc.dealloc(ptr, layout) }
@@ -164,8 +165,7 @@ where
         // SAFETY: `ptr` is at least `layout.align()` byte aligned and this is a power
         // of two
         unsafe {
-            self.zeroizer
-                .zeroize_mem_minaligned(ptr.as_ptr(), layout.size(), layout.align());
+            self.zeroizer.zeroize_mem(ptr.as_ptr(), layout.size());
         }
         // SAFETY: caller must uphold the safety contract of `Allocator::deallocate`
         unsafe { self.backend_alloc.deallocate(ptr, layout) }
